#include <libc.h>
#include <dirent.h>
#include <sys/mman.h>
#include <sys/shobj.h>


int verbose = 0;

dump(unsigned char *p, int len, FILE *f)
{
	int	i;
	for (i=0; i < len; i += 16) {
		int	j;
		for (j=0; j < 16 && j+i < len; j++) {
			fprintf(f,"%02.02x ",p[i+j]);
		}
		for (j=0; j < 16 && j+i < len; j++) {
			putc(isprint(p[i+j]) ? p[i+j] : '.', f);
		}
		putc('\n',f);
	}
}

struct flags {
	int	flag;
	char    *str;
};

print_flags(int fl, struct flags *fp, int N, FILE *f)
{
	while (N--) {
		if (fl & fp->flag) {
			fputs(fp->str, f);
		} else {
			putc('-', f);
		}
		fp++;
	}
}

struct flags prot_flags[] = {
	PROT_READ,	"r",
	PROT_WRITE,	"w",
	PROT_EXEC,	"x",
	PROT_NOCACHE,	"v"
};

struct flags map_flags[] = {
	MAP_SHARED,     "s",
	MAP_PRIVATE,	"p",
	MAP_FIXED,	"f"
};


print_shobj(char *name, FILE *f)
{
	int	fd;
	struct	s_bind	*sp;
	int	i;
	if ((fd=shm_open(name,O_RDONLY, 0)) == -1) {
		fprintf(stderr,"shm_open failed '%s' (%s)\n",
			name, strerror(errno));
		return -1;
	}
	if ((sp=mmap(0,sizeof *sp, PROT_READ, MAP_SHARED, fd, 0)) == (void*)-1) {
		perror("mmap");
		close(fd);
	}
	if (sp->s_magic != SHO_MAGIC) {
		if (verbose) {
			fprintf(f, "%s is not a shared memory object\n", name);
		}
		munmap(sp, sizeof *sp);
		return 1;
	}
	fprintf(f,"%s (%s) version %u; %u\n",
		name, sp->s_regnam, sp->s_version);
	for (i=0; i < sp->s_nseg; i++) {
		fprintf(f, " seg %u ", i);
		print_flags(sp->s_segs[i].seg_prot, prot_flags,
			sizeof prot_flags/sizeof *prot_flags, f);
		putc(',', f);
		print_flags(sp->s_segs[i].seg_map, map_flags,
			sizeof map_flags/sizeof *map_flags, f);
		fprintf(f," [%-08.8x..%-08.8x]  at %-08.8x",
			i, sp->s_segs[i].seg_addr,
			(caddr_t)sp->s_segs[i].seg_addr +
				sp->s_segs[i].seg_len -1,
			sp->s_segs[i].seg_offs);
		putc('\n', f);
		if (verbose) {
			unsigned char *pmem;
			pmem = mmap(0,sp->s_segs[i].seg_len,
					PROT_READ, MAP_SHARED, fd,
					sp->s_segs[i].seg_offs);
			if (pmem == (void *)-1) {
				fprintf(stderr,"error: cannot map (%s)\n",
					strerror(errno));
			} else {
				dump(pmem, sp->s_segs[i].seg_len, f);
				munmap(pmem, sp->s_segs[i].seg_len);
			}
		}
	}
//	fputs("links\n", f);
	for (i=0; i < sp->s_nlnk; i++) {
		fprintf(f," link %u: %-08.8x..%-08.8x\n",
			i,
			sp->s_lnks[i].lnk_addr,
			(caddr_t)sp->s_lnks[i].lnk_addr+sp->s_lnks[i].lnk_len-1);
	}
	munmap(sp, sizeof *sp);
	return 0;
}

main(int argc, char **argv)
{
	int	c;
	while ((c=getopt(argc,argv,"v")) != EOF) {
		switch (c) {
		case 'v': verbose = 1; break;
		default:  return 1;
		}
	}
	if (optind == argc) {
		DIR *d = opendir("/dev/shmem");
		struct dirent *p;
		while (p=readdir(d)) {
			print_shobj(p->d_name, stdout);
		}
		closedir(d);
	} else while (optind < argc) {
		print_shobj(argv[optind++], stdout);
	}
	return 0;
}
